#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#define DAYTIME_PORT  13   // Port of daytime service (RFC 867)
#define BUFFER_SIZE   100  // Size of buffer for daytime string

/* Name:         getCmdLineArgs
 * Description:  Gets the expected arguments from command line
 * Parameters:   argc - number of arguments specified
 *               argv - array with arguments as strings
 *               server - hostname of server to connect to
 * Return code:  true if successful
 *               false if an error occurred
 */
bool getCmdLineArgs(int argc, char *argv[], const char **server)
{
	bool result = false;

	if (argc == 2)
	{
		*server = argv[1];
		result = true;
	}
	else
	{
		fprintf(stderr, "usage: %s <server>\n", argv[0]);
	}

	return result;
}

/* Name:         resolveHostName
 * Description:  Resolves a hostname into an IP address
 * Parameters:   name - hostname to resolve
 *               addr - buffer to store IP address in
 * Return code:  true if successful
 *               false if an error occurred
 */
bool resolveHostName(const char *name, struct in_addr *addr)
{
	bool result = false;

	printf("Resolving hostname \"%s\"...", name);
	struct hostent *host = gethostbyname(name);
	if (host)
	{
		printf(" OK.\n");
		*addr = *(struct in_addr*)host->h_addr;
		result = true;
	}
	else
	{
		printf(" FAILED.\n");
		fprintf(stderr, "Couldn't resolve hostname: %s\n", hstrerror(h_errno));
	}

	return result;
}

/* Name:         createTcpSocket
 * Description:  Creates a new, unconnected TCP socket
 * Parameters:   sock - buffer to receive socket
 * Return code:  true if successful
 *               false if an error occurred
 */
bool createTcpSocket(int *sock)
{
	bool failed = false;

	*sock = socket(AF_INET, SOCK_STREAM, 0);
	if (*sock == -1)
	{
		fprintf(stderr, "Couldn't create socket: %s\n", strerror(errno));
		failed = true;
	}

	return !failed;
}

/* Name:         connectTcpSocket
 * Description:  Connects a TCP socket to a destination end point
 * Parameters:   sock - socket to connect
 *               addr - destination address to connect to
 *               port - destination port to connect to
 * Return code:  true if successful
 *               false if an error occurred
 */
bool connectTcpSocket(int sock, struct in_addr addr, unsigned short port)
{
	bool result = false;

	printf("Connecting to %s:%u...", inet_ntoa(addr), port);
	fflush(stdout);

	struct sockaddr_in saddr;
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	saddr.sin_addr = addr;
	if (connect(sock, (struct sockaddr*) &saddr, sizeof(saddr)) == 0)
	{
		printf(" OK.\n");
		result = true;
	}
	else
	{
		printf(" FAILED.\n");
		fprintf(stderr, "Couldn't connect to host: %s\n", strerror(errno));
	}

	return result;
}

/* Name:         receiveString
 * Description:  Receives and prints a string from a connected socket
 * Parameters:   sock - socket to read from
 * Return code:  true if successful
 *               false if an error occurred
 */
bool receiveString(int sock)
{
	int total = 0;
	bool done = false;
	bool failed = false;
	char buffer[BUFFER_SIZE];

	while (!done && !failed)
	{
		int bytes = recv(sock, buffer + total, sizeof(buffer) - total - 1, 0);
		if (bytes > 0)
		{
			total += bytes;
			buffer[total] = '\0';
		}
		else if (bytes == 0)
		{
			done = true;
		}
		else if (bytes == -1)
		{
			fprintf(stderr, "Couldn't read from socket: %s\n", strerror(errno));
			failed = true;
		}
	}

	if (!failed)
	{
		printf("Received %i bytes: <%s>\n", total, buffer);
	}

	return !failed;
}

/* Name:         safeCloseSocket
 * Description:  Checks if a socket is valid, closes it. The socket is
 *               set to an invalid value to avoid double free errors.
 * Parameters:   sock - socket to communicate with
 * Return code:  none
 */
void safeCloseSocket(int *sock)
{
	if (*sock != -1)
	{
		close(*sock);
		*sock = -1;
	}
}

int main(int argc, char *argv[])
{
	const char *hostname;
	bool good = getCmdLineArgs(argc, argv, &hostname);

	struct in_addr addr;
	if (good)
		good = resolveHostName(hostname, &addr);

	int sock = -1;
	if (good)
		good = createTcpSocket(&sock);

	if (good)
		good = connectTcpSocket(sock, addr, DAYTIME_PORT);

	if (good)
		good = receiveString(sock);

	safeCloseSocket(&sock);

	return good ? EXIT_SUCCESS : EXIT_FAILURE;
}
